package org.artifactory.workqueue;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.maven.MavenMetadataService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.async.AsyncExecutor;
import org.artifactory.build.InternalBuildService;
import org.artifactory.common.ConstantValues;
import org.artifactory.concurrent.ArtifactoryRunnable;
import org.artifactory.mbean.MBeanRegistrationService;
import org.artifactory.security.SystemAuthenticationToken;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.storage.jobs.migration.buildinfo.BuildInfoMigrationJob;
import org.artifactory.storage.jobs.migration.pathchecksum.RepoPathChecksumMigrationJob;
import org.artifactory.storage.jobs.migration.sha256.Sha256MigrationJob;
import org.jetbrains.annotations.NotNull;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;

import java.util.List;

@SuppressWarnings("SpringJavaInjectionPointsAutowiringInspection")
@Configuration
public class WorkQueueConfig implements ContextCreationListener {

    private final WorkQueueConflictsGuardProvider defaultConflictsGuardProvider;

    public WorkQueueConfig(@NotNull AddonsManager addonsManager) {
        this.defaultConflictsGuardProvider =
                (key, info) -> addonsManager.addonByType(HaAddon.class).getConflictsGuard(key);
    }

    @Bean
    public AsyncWorkQueueService workQueueService(@NotNull List<WorkQueueInfo> workQueueInfos,
            @NotNull AsyncExecutor defaultExecutor,
            @NotNull List<RemoveWorkQueuePredicate> removeWorkQueuePredicates,
            @NotNull MBeanRegistrationService mBeanRegistrationService) {
        return new AsyncWorkQueueServiceImpl(workQueueInfos, defaultExecutor, removeWorkQueuePredicates,
                mBeanRegistrationService,
                ConstantValues.workQueueMaxExecutionSec.getInt(),
                ConstantValues.workQueueSyncExecutionTimeoutSec.getInt(),
                ConstantValues.workQueueJobEnabled.getBoolean() ? ConstantValues.workQueueDoJobIntervalSecs.getInt() :
                        0);
    }

    @Override
    public void onContextCreated() {
        ArtifactoryContext context = ContextHelper.get();
        ContextHelper.get().beanForType(AsyncWorkQueueService.class).scheduleWorkQueueCleanupJob(
                delegate -> new ArtifactoryRunnable(delegate, context, new SystemAuthenticationToken()));
    }

    @Bean
    public WorkQueueInfo mavenMetadataWorkQueueInfo() {
        return new WorkQueueInfo(ConstantValues.mvnMetadataCalculationWorkers.getInt(),
                MavenMetadataService.MAVEN_METADATA_WQ_NAME, defaultConflictsGuardProvider);
    }

    @Bean
    public WorkQueueInfo mavenPluginsMetadataWorkQueueInfo() {
        return new WorkQueueInfo(ConstantValues.mvnMetadataPluginCalculationWorkers.getInt(),
                MavenMetadataService.MAVEN_PLUGIN_METADATA_WQ_NAME, defaultConflictsGuardProvider);
    }

    @Bean
    public WorkQueueInfo buildRetentionWorkQueueInfo() {
        return new WorkQueueInfo(ConstantValues.buildRetentionWorkers.getInt(),
                InternalBuildService.BUILD_RETENTION_JOB_WQ_NAME, defaultConflictsGuardProvider);
    }

    @Bean
    public WorkQueueInfo sha2MigrationWorkQueueInfo() {
        return new WorkQueueInfo(ConstantValues.sha2MigrationJobQueueWorkers.getInt(),
                Sha256MigrationJob.WORK_QUEUE_NAME, defaultConflictsGuardProvider);
    }

    @Bean
    public WorkQueueInfo buildInfoMigrationWorkQueueInfo() {
        return new WorkQueueInfo(ConstantValues.buildInfoMigrationJobQueueWorkers.getInt(),
                BuildInfoMigrationJob.WORK_QUEUE_NAME, defaultConflictsGuardProvider);
    }

    @Bean
    public WorkQueueInfo pathChecksumMigrationWorkQueueInfo() {
        return new WorkQueueInfo(ConstantValues.pathChecksumMigrationJobQueueWorkers.getInt(),
                RepoPathChecksumMigrationJob.WORK_QUEUE_NAME, defaultConflictsGuardProvider);
    }

    @Bean
    public WorkQueueInfo deletePathItemWorkQueueInfo() {
        return new WorkQueueInfo(ConstantValues.asyncDeleteWorkers.getInt(),
                RepositoryService.ASYNC_DELETE_PATH_ITEM_WQ_NAME, defaultConflictsGuardProvider);
    }
}
