package org.artifactory.queue.behaviors;

import org.artifactory.storage.event.EventInfo;

import java.util.List;

import static com.google.common.collect.Lists.newArrayList;
import static org.artifactory.storage.event.EventType.create;
import static org.artifactory.storage.event.EventType.delete;
import static org.artifactory.storage.event.EventType.props;
import static org.artifactory.storage.event.EventType.update;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertThrows;

/**
 * @author avishayh
 */
public class EliminateEventBehaviorTest extends AbstractEventReductionBehaviorTest {

    public void deleteEliminateUpdateOrCreateBehaviour() {
        EliminateEventBehavior behavior = new DeleteEliminateUpdateOrCreateBehavior();
        List<EventInfo> input;
        List<EventInfo> output;

        //null doesnt fail
        output = behavior.apply(null);
        assertNull(output, "null does not fail");
        //input equals output
        input = newArrayList();
        add(input, create, "a");
        output = behavior.apply(input);
        assertEquals(output, input, "no eliminations here");
        //input equals output
        input = newArrayList();
        add(input, create, "a");
        add(input, update, "a");
        output = behavior.apply(input);
        assertEquals(output, input, "no eliminations here");
        //input equals output
        input = newArrayList();
        add(input, delete, "a");
        add(input, delete, "a");
        output = behavior.apply(input);
        assertEquals(output, input, "no eliminations here");
        //eliminating all events
        input = newArrayList();
        add(input, create, "a"); //eliminated
        add(input, delete, "a"); //eliminated
        output = behavior.apply(input);
        assertEquals(output.size(), 0, "eliminating all events");
        input = newArrayList();
        add(input, update, "a"); //eliminated
        add(input, delete, "a"); //eliminated
        output = behavior.apply(input);
        assertEquals(output.size(), 0, "eliminating all events");
        input = newArrayList();
        add(input, create, "a"); //eliminated
        add(input, update, "a"); //eliminated
        add(input, delete, "a"); //eliminated
        output = behavior.apply(input);
        assertEquals(output.size(), 0, "eliminating all events");

        //new create event survives
        input = newArrayList();
        add(input, create, "a"); //eliminated
        add(input, update, "a"); //eliminated
        add(input, delete, "a"); //eliminated
        add(input, create, "a");
        int currentIndex = index;
        output = behavior.apply(input);
        assertEquals(output.size(), 1, "same as before with extra event survived after");
        assertEquals(output.get(0).getEventId(), currentIndex, "same as before with extra event survived after");

        //now with multiple path + survivors
        input = newArrayList();
        add(input, create, "a"); //eliminated
        add(input, create, "b"); //eliminated
        add(input, update, "b"); //eliminated
        add(input, create, "c");
        int cIndex = index;
        add(input, update, "a"); //eliminated
        add(input, delete, "b"); //eliminates (create,update b) + self
        add(input, delete, "a"); //eliminates (create,update a) + self
        add(input, create, "a");
        int aIndex = index;
        output = behavior.apply(input);
        assertEquals(output.size(), 2, "multiple path same as before");
        assertEquals(output.get(0).getEventId(), cIndex, "multiple path same as before");
        assertEquals(output.get(1).getEventId(), aIndex, "multiple path same as before");
    }

    public void eliminateDedupBehaviorWithoutSelfElimination() {
        CreateEliminateDeleteBehavior behavior = new CreateEliminateDeleteBehavior();
        List<EventInfo> input;
        List<EventInfo> output;

        input = newArrayList();
        add(input, delete, "a"); //eliminated
        add(input, create, "a");
        output = behavior.apply(input);
        assertEquals(output.size(), 1, "create eliminates delete");
        assertEquals(output.get(0).getType(), create, "create eliminates delete");

        reset();
        input = newArrayList();
        add(input, props, "a");
        add(input, delete, "a"); //eliminated
        add(input, create, "b");
        add(input, create, "a");
        add(input, props, "a");
        output = behavior.apply(input);
        assertEquals(output.size(), 4);
        assertEquals(output.get(0).toString(), "1|props|a");
        assertEquals(output.get(1).toString(), "3|create|b");
        assertEquals(output.get(2).toString(), "4|create|a");
        assertEquals(output.get(3).toString(), "5|props|a");
    }

    public void illegalEliminatorBehaviour() {
        assertThrows(IllegalArgumentException.class, () -> new EliminateEventBehavior().apply(newArrayList()));
        assertThrows(IllegalArgumentException.class,
                () -> new EliminateEventBehavior().eliminator(create).apply(newArrayList()));
        assertThrows(IllegalArgumentException.class,
                () -> new EliminateEventBehavior().eliminator(create).eliminates(create, update));
        assertThrows(IllegalArgumentException.class, () -> new EliminateEventBehavior().eliminates(create, update));
    }
}
