package org.artifactory.security.authentication;

import org.apache.http.HttpHeaders;
import org.artifactory.addon.oauth.OAuthHandler;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.security.SecurityDescriptor;
import org.artifactory.descriptor.security.oauth.OAuthSettings;
import org.artifactory.request.RequestThreadLocal;
import org.artifactory.request.RequestWrapper;
import org.artifactory.security.UserInfo;
import org.artifactory.security.props.auth.DockerTokenManager;
import org.artifactory.security.props.auth.model.AuthenticationModel;
import org.artifactory.security.props.auth.model.OauthErrorEnum;
import org.artifactory.security.props.auth.model.OauthErrorModel;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.spring.InternalContextHelper;
import org.jetbrains.annotations.NotNull;
import org.mockito.Mockito;
import org.springframework.security.core.Authentication;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.Test;

import javax.servlet.http.HttpServletRequest;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

/**
 * @author Alexei Vainshtein
 */
@Test
public class GitHubEnterpriseAuthenticationProviderTest {

    public void testAuthenticateWithoutSuccess() {
        UserGroupService userGroupService = Mockito.mock(UserGroupService.class);
        GitHubEnterpriseAuthenticationProvider provider =
                new GitHubEnterpriseAuthenticationProvider(null, userGroupService);
        assertNull(provider.authenticate(null));

        Authentication authentication = createAuthentication();
        bindRequest("really/docker/v2/token");
        CentralConfigDescriptor descriptor = bindMocks();
        SecurityDescriptor securityDescriptor = bindSecurity(descriptor);
        OAuthSettings oAuthSettings = bindOauth(securityDescriptor);
        oAuthSettings.setEnableIntegration(true);
        oAuthSettings.setDefaultNpm("default");
        OAuthHandler oAuthHandler = Mockito.mock(OAuthHandler.class);
        Mockito.when(ContextHelper.get().beanForType(OAuthHandler.class)).thenReturn(oAuthHandler);
        OauthErrorModel errorModel = createErrorModel();
        Mockito.when(oAuthHandler.getCreateToken(Mockito.anyString(), Mockito.eq(null))).thenReturn(errorModel);
        assertNull(provider.authenticate(authentication));
    }

    public void testAuthenticateWithSuccess() {
        UserGroupService userGroupService = Mockito.mock(UserGroupService.class);
        GitHubEnterpriseAuthenticationProvider provider =
                new GitHubEnterpriseAuthenticationProvider(null, userGroupService);

        Authentication authentication = createAuthentication();
        bindRequest("really/docker/v2/token");
        CentralConfigDescriptor descriptor = bindMocks();
        SecurityDescriptor securityDescriptor = bindSecurity(descriptor);
        OAuthSettings oAuthSettings = bindOauth(securityDescriptor);
        oAuthSettings.setEnableIntegration(true);
        oAuthSettings.setDefaultNpm("default");
        OAuthHandler oAuthHandler = Mockito.mock(OAuthHandler.class);
        Mockito.when(ContextHelper.get().beanForType(OAuthHandler.class)).thenReturn(oAuthHandler);
        AuthenticationModel model = new AuthenticationModel("token", "at");

        UserInfo userInfo = Mockito.mock(UserInfo.class);
        Mockito.when(userGroupService
                .findOrCreateExternalAuthUser(Mockito.anyString(), Mockito.anyBoolean(), Mockito.anyBoolean()))
                .thenReturn(userInfo);
        Mockito.when(oAuthHandler.getCreateToken(Mockito.anyString(), Mockito.eq(null)))
                .thenReturn(model);
        DockerTokenManager dockerTokenManager = Mockito.mock(DockerTokenManager.class);
        Mockito.when(InternalContextHelper.get().beanForType(DockerTokenManager.class)).thenReturn(dockerTokenManager);
        Authentication authenticate = provider.authenticate(authentication);
        assertNotNull(authenticate);
        assertTrue(authenticate instanceof OauthAuthentication);
    }

    public void testSupportsNullRequest() {
        RequestWrapper requestWrapper = Mockito.mock(RequestWrapper.class);
        RequestThreadLocal.bind(requestWrapper);
        GitHubEnterpriseAuthenticationProvider provider = new GitHubEnterpriseAuthenticationProvider(null, null);
        assertFalse(provider.supports(null));
    }

    public void testSupportsNonV2Request() {
        bindRequest("really/not/docker");
        GitHubEnterpriseAuthenticationProvider provider = new GitHubEnterpriseAuthenticationProvider(null, null);
        assertFalse(provider.supports(null));
    }

    public void testSupportsV2RequestNoBasicAuth() {
        bindRequest("api/docker/v2/token");
        GitHubEnterpriseAuthenticationProvider provider = new GitHubEnterpriseAuthenticationProvider(null, null);
        assertFalse(provider.supports(null));
    }

    public void testSupportsV2RequestNoApiDocker() {
        bindRequest("really/docker/v2/token");
        GitHubEnterpriseAuthenticationProvider provider = new GitHubEnterpriseAuthenticationProvider(null, null);
        assertFalse(provider.supports(null));
    }

    public void testSupportsV2RequestNoOauthEnabled() {
        HttpServletRequest request = bindRequest("really/docker/v2/token");
        Mockito.when(request.getHeader(HttpHeaders.AUTHORIZATION)).thenReturn("Basic auto");

        CentralConfigDescriptor descriptor = bindMocks();
        SecurityDescriptor securityDescriptor = bindSecurity(descriptor);

        GitHubEnterpriseAuthenticationProvider provider = new GitHubEnterpriseAuthenticationProvider(null, null);
        assertFalse(provider.supports(null));

        bindOauth(securityDescriptor);
        assertFalse(provider.supports(null));
    }

    public void testSupportsV2RequestOauthEnabled() {
        HttpServletRequest request = bindRequest("api/docker/v2/token");
        Mockito.when(request.getHeader(HttpHeaders.AUTHORIZATION)).thenReturn("Basic auto");

        CentralConfigDescriptor descriptor = bindMocks();
        SecurityDescriptor securityDescriptor = bindSecurity(descriptor);

        GitHubEnterpriseAuthenticationProvider provider = new GitHubEnterpriseAuthenticationProvider(null, null);
        assertFalse(provider.supports(null));

        OAuthSettings oAuthSettings = bindOauth(securityDescriptor);
        oAuthSettings.setEnableIntegration(true);
        assertTrue(provider.supports(null));
    }

    public void testGetRealm() {
        GitHubEnterpriseAuthenticationProvider provider = new GitHubEnterpriseAuthenticationProvider(null, null);
        assertEquals(provider.getRealm(), OAuthHandler.REALM);
    }

    @NotNull
    private OAuthSettings bindOauth(SecurityDescriptor securityDescriptor) {
        OAuthSettings oAuthSettings = new OAuthSettings();
        Mockito.when(securityDescriptor.getOauthSettings()).thenReturn(oAuthSettings);
        return oAuthSettings;
    }

    private SecurityDescriptor bindSecurity(CentralConfigDescriptor descriptor) {
        SecurityDescriptor securityDescriptor = Mockito.mock(SecurityDescriptor.class);
        Mockito.when(descriptor.getSecurity()).thenReturn(securityDescriptor);
        return securityDescriptor;
    }

    private CentralConfigDescriptor bindMocks() {
        ArtifactoryContext context = Mockito.mock(InternalArtifactoryContext.class);
        CentralConfigService configService = Mockito.mock(CentralConfigService.class);
        CentralConfigDescriptor descriptor = Mockito.mock(CentralConfigDescriptor.class);
        Mockito.when(configService.getDescriptor()).thenReturn(descriptor);
        Mockito.when(context.getCentralConfig()).thenReturn(configService);
        ArtifactoryContextThreadBinder.bind(context);
        return descriptor;
    }

    @NotNull
    private OauthErrorModel createErrorModel() {
        OauthErrorModel errorModel = new OauthErrorModel();
        errorModel.setInternalErrorMsg(OauthErrorEnum.BAD_REQUEST);
        errorModel.setStatusCode(400);
        return errorModel;
    }

    private HttpServletRequest bindRequest(String uri) {
        RequestWrapper requestWrapper = Mockito.mock(RequestWrapper.class);
        HttpServletRequest request = Mockito.mock(HttpServletRequest.class);
        Mockito.when(requestWrapper.getRequest()).thenReturn(request);
        Mockito.when(request.getRequestURI()).thenReturn(uri);
        RequestThreadLocal.bind(requestWrapper);
        return request;
    }

    @NotNull
    private Authentication createAuthentication() {
        Authentication authentication = Mockito.mock(Authentication.class);
        Mockito.when(authentication.getPrincipal()).thenReturn("test");
        Mockito.when(authentication.getCredentials()).thenReturn("test");
        return authentication;
    }

    @AfterMethod
    public void unbindRequest() {
        RequestThreadLocal.unbind();
        ArtifactoryContextThreadBinder.unbind();
    }
}