package org.artifactory.servicetrust.cold;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.retention.ArtifactRetentionAddon;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.access.GenericTokenSpec;
import org.artifactory.api.security.access.TokenSpec;
import org.artifactory.servicetrust.ExchangeTokenRequest;
import org.artifactory.servicetrust.ExchangeTokenResponse;
import org.artifactory.servicetrust.ServiceTrust;
import org.artifactory.servicetrust.ServiceTrustUseCase;
import org.artifactory.servicetrust.TokenExtension;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.token.JwtAccessToken;
import org.jfrog.common.JsonUtils;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.security.core.context.SecurityContextHolder;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.List;

import static org.artifactory.servicetrust.ServiceTrustService.DEFAULT_EXCHANGE_TOKEN_EXPIRATION_TIME_IN_SECS;
import static org.artifactory.servicetrust.ServiceTrustService.DEFAULT_MASTER_TOKEN_EXPIRATION_TIME_IN_SECS;
import static org.artifactory.servicetrust.ServiceTrustService.DEFAULT_PAIRING_TOKEN_EXPIRATION_TIME_IN_SECS;
import static org.artifactory.servicetrust.ServiceTrustService.INTERNAL_SCOPE_PREFIX;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertFalse;
import static org.testng.AssertJUnit.assertTrue;

public class ArtifactoryColdTokenSpecServiceTest extends ArtifactoryHomeBoundTest {
    private static final String BASE_URL = "baseUrl";
    private static final String EXCHANGE_URL = "exchangeUrl";
    private static final String PAIRING_URL = "pairingUrl";
    private static final String SUBJECT = "a@b/users/username";
    private static final ServiceId SERVICE_ID = new ServiceId("artifactory", "serviceId");
    private static final String TOKEN_ID = "tokenId";
    private static final String ACCESS_TOKEN = "bearer_access_token";
    private static final String NAMESPACE = "namespace";
    private static final long EXPIRES_IN = 10000;
    private static final String SCOPE_WITH_NAMESPACE =
            "internal:service-trust/pairing/artifactory-cold/" + NAMESPACE + ":x";
    private static final String SCOPE_WITHOUT_NAMESPACE = "internal:service-trust/pairing/artifactory-cold:x";
    private static final String NON_COLD_SCOPE = "internal:service-trust/pairing/artifactory-federation:x";
    private static final String COLD_USE_CASE = "artifactory-cold/namespace";
    private static final String LIMITED_COLD_USE_CASE = "artifactory-cold/namespace:scoped";

    @Mock
    private InternalArtifactoryContext context;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private ArtifactRetentionAddon artifactRetentionAddon;

    private AutoCloseable closeableMock;

    @BeforeMethod
    private void setup() {
        closeableMock = MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(context);
        when(addonsManager.addonByType(ArtifactRetentionAddon.class)).thenReturn(artifactRetentionAddon);
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
    }

    @AfterMethod
    private void cleanup() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        SecurityContextHolder.clearContext();
        closeableMock.close();
    }

    @Test
    public void testCreatePairingTokenSpecWithoutNamespace() {
        when(artifactRetentionAddon.generateJFrogColdStorageNamespace()).thenReturn(NAMESPACE);
        when(artifactRetentionAddon.isColdInstance()).thenReturn(true);

        TokenSpec<GenericTokenSpec> tokenSpec = ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE
                .createPairingTokenSpec(SERVICE_ID, ServiceTrust.ONE_WAY, SUBJECT, SCOPE_WITHOUT_NAMESPACE, BASE_URL,
                        PAIRING_URL, EXCHANGE_URL);

        assertEquals(1, tokenSpec.getAudience().size());
        assertTrue(tokenSpec.getAudience().contains(SERVICE_ID.toString()));

        List<String> scopeList = tokenSpec.getScope();
        assertEquals(1, scopeList.size());
        assertTrue(scopeList.contains(SCOPE_WITH_NAMESPACE));

        assertFalse(tokenSpec.getRefreshable());
        assertEquals(DEFAULT_PAIRING_TOKEN_EXPIRATION_TIME_IN_SECS, tokenSpec.getExpiresIn().longValue());
        assertEquals(buildExtensionJson(), tokenSpec.getExtension());
    }

    @Test
    public void testCreatePairingTokenSpecWithNamespace() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(true);

        TokenSpec<GenericTokenSpec> tokenSpec = ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE
                .createPairingTokenSpec(SERVICE_ID, ServiceTrust.ONE_WAY, SUBJECT, SCOPE_WITH_NAMESPACE,
                        BASE_URL,
                        PAIRING_URL,
                        EXCHANGE_URL);

        assertEquals(1, tokenSpec.getAudience().size());
        assertTrue(tokenSpec.getAudience().contains(SERVICE_ID.toString()));

        List<String> scopeList = tokenSpec.getScope();
        assertEquals(1, scopeList.size());
        assertTrue(scopeList.contains(SCOPE_WITH_NAMESPACE));

        assertFalse(tokenSpec.getRefreshable());
        assertEquals(DEFAULT_PAIRING_TOKEN_EXPIRATION_TIME_IN_SECS, tokenSpec.getExpiresIn().longValue());
        assertEquals(buildExtensionJson(), tokenSpec.getExtension());
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "artifactory-cold " +
            "does not support two-way pairing")
    public void testCreatePairingTokenSpecWithTwoWayTrust() {
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.createPairingTokenSpec(SERVICE_ID,
                ServiceTrust.TWO_WAY, SUBJECT,
                "scope", BASE_URL, PAIRING_URL, EXCHANGE_URL);
    }

    @Test(expectedExceptions = IllegalStateException.class, expectedExceptionsMessageRegExp =
            "Cold Storage mode is disabled")
    public void testCreatePairingTokenSpecForNonColdInstance() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.createPairingTokenSpec(SERVICE_ID,
                ServiceTrust.ONE_WAY, SUBJECT,
                "scope", BASE_URL, PAIRING_URL, EXCHANGE_URL);
    }

    @Test
    public void testCreateMasterTokenSpec() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(true);

        String pairingScope = "internal:service-trust/pairing/" + COLD_USE_CASE + ":x";
        TokenSpec<GenericTokenSpec> tokenSpec = ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE
                .createMasterTokenSpec(SERVICE_ID, COLD_USE_CASE, pairingScope, BASE_URL, PAIRING_URL, EXCHANGE_URL);

        assertEquals(1, tokenSpec.getAudience().size());
        assertTrue(tokenSpec.getAudience().contains(SERVICE_ID.toString()));

        String adminScope = SERVICE_ID + ":admin";
        String requestedScope = INTERNAL_SCOPE_PREFIX + COLD_USE_CASE;

        List<String> scopeList = tokenSpec.getScope();
        assertEquals(3, scopeList.size());
        assertTrue(scopeList.contains(pairingScope));
        assertTrue(scopeList.contains(adminScope));
        assertTrue(scopeList.contains(requestedScope));

        assertTrue(tokenSpec.getRefreshable());
        assertEquals(DEFAULT_MASTER_TOKEN_EXPIRATION_TIME_IN_SECS, tokenSpec.getExpiresIn().longValue());
        assertEquals(buildExtensionJson(), tokenSpec.getExtension());
    }

    @Test(expectedExceptions = IllegalStateException.class, expectedExceptionsMessageRegExp =
            "Cold Storage mode is disabled")
    public void testCreateMasterTokenSpecForNonColdInstance() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.createMasterTokenSpec(SERVICE_ID, "useCase",
                "pairingScope", BASE_URL, PAIRING_URL, EXCHANGE_URL);
    }

    @Test
    public void testExchangeTokenSpec() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(true);
        String useCase = "artifactory-cold/namespace:scoped " + SERVICE_ID + ":admin";

        TokenSpec<GenericTokenSpec> tokenSpec = ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE
                .createExchangeTokenSpec(SERVICE_ID, useCase);

        assertEquals(1, tokenSpec.getAudience().size());
        assertTrue(tokenSpec.getAudience().contains(SERVICE_ID.toString()));

        String adminScope = SERVICE_ID + ":admin";
        String requestedScope = INTERNAL_SCOPE_PREFIX + useCase;

        List<String> scopeList = tokenSpec.getScope();
        assertEquals(2, scopeList.size());
        assertTrue(scopeList.contains(adminScope));
        assertTrue(scopeList.contains(requestedScope));

        assertFalse(tokenSpec.getRefreshable());
        assertEquals(DEFAULT_EXCHANGE_TOKEN_EXPIRATION_TIME_IN_SECS, tokenSpec.getExpiresIn().longValue());
        assertEquals(buildOnlyRevocableExtensionJson(), tokenSpec.getExtension());
    }

    @Test(expectedExceptions = IllegalStateException.class, expectedExceptionsMessageRegExp =
            "Cold Storage mode is disabled")
    public void testExchangeTokenSpecForNonColdInstance() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.createExchangeTokenSpec(SERVICE_ID,
                "useCase");
    }

    @Test
    public void testCreateExchangeTokenRequestWithNamespace() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        JwtAccessToken jwtAccessToken = mock(JwtAccessToken.class);
        when(jwtAccessToken.getScope()).thenReturn(List.of(SCOPE_WITH_NAMESPACE));
        ExchangeTokenRequest exchangeTokenRequest = ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE
                .createExchangeTokenRequest(SUBJECT, jwtAccessToken);
        assertEquals(SUBJECT, exchangeTokenRequest.getSubject());
        assertEquals(COLD_USE_CASE, exchangeTokenRequest.getTarget());
    }

    @Test
    public void testCreateExchangeTokenRequestWithoutNamespace() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        when(artifactRetentionAddon.getJfrogColdStorageArchivingNamespace()).thenReturn(NAMESPACE);
        JwtAccessToken jwtAccessToken = mock(JwtAccessToken.class);
        when(jwtAccessToken.getScope()).thenReturn(List.of(SCOPE_WITHOUT_NAMESPACE));
        ExchangeTokenRequest exchangeTokenRequest = ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE
                .createExchangeTokenRequest(SUBJECT, jwtAccessToken);
        assertEquals(SUBJECT, exchangeTokenRequest.getSubject());
        assertEquals(ServiceTrustUseCase.ARTIFACTORY_COLD.getName(), exchangeTokenRequest.getTarget());
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "Pair's token scope " +
            "can not be null")
    public void testCreateExchangeTokenRequestWhenJwtScopeNull() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        JwtAccessToken jwtAccessToken = mock(JwtAccessToken.class);
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.createExchangeTokenRequest(SUBJECT,
                jwtAccessToken);
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "Pair's token does " +
            "not have retention related scope")
    public void testCreateExchangeTokenRequestWithNonColdJwtScope() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        JwtAccessToken jwtAccessToken = mock(JwtAccessToken.class);
        when(jwtAccessToken.getScope()).thenReturn(List.of(NON_COLD_SCOPE));
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.createExchangeTokenRequest(SUBJECT,
                jwtAccessToken);
    }

    @Test
    public void testStoreMasterTokenData() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        ExchangeTokenResponse exchangeTokenResponse =
                new ExchangeTokenResponse(TOKEN_ID, ACCESS_TOKEN, EXPIRES_IN, SCOPE_WITH_NAMESPACE);
        TokenExtension tokenExtension = new TokenExtension(BASE_URL, EXCHANGE_URL, null, true);

        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.storeMasterTokenData(exchangeTokenResponse,
                tokenExtension);

        ArgumentCaptor<String> baseUrlCapture = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<String> exchangeUrlCapture = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<String> namespaceUrlCapture = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<String> tokenIdCapture = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<String> accessTokenCapture = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<String> tokenTypeCapture = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<Long> expireInCapture = ArgumentCaptor.forClass(Long.class);
        ArgumentCaptor<String> scopeCapture = ArgumentCaptor.forClass(String.class);

        verify(artifactRetentionAddon, times(1))
                .saveNamespaceAndMasterToken(baseUrlCapture.capture(),
                        exchangeUrlCapture.capture(),
                        namespaceUrlCapture.capture(),
                        tokenIdCapture.capture(),
                        accessTokenCapture.capture(),
                        tokenTypeCapture.capture(),
                        expireInCapture.capture(),
                        scopeCapture.capture());
        assertEquals(baseUrlCapture.getValue(), BASE_URL);
        assertEquals(exchangeUrlCapture.getValue(), EXCHANGE_URL);
        assertEquals(namespaceUrlCapture.getValue(), NAMESPACE);
        assertEquals(tokenIdCapture.getValue(), TOKEN_ID);
        assertEquals(accessTokenCapture.getValue(), ACCESS_TOKEN);
        assertEquals(tokenTypeCapture.getValue(), "access_token");
        assertEquals(expireInCapture.getValue().longValue(), EXPIRES_IN);
        assertEquals(scopeCapture.getValue(), SCOPE_WITH_NAMESPACE);
    }

    @Test(expectedExceptions = IllegalArgumentException.class,
            expectedExceptionsMessageRegExp = "Master token does not have scope")
    public void testStoreMasterTokenDataWhenScopeIsNull() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        ExchangeTokenResponse exchangeTokenResponse =
                new ExchangeTokenResponse(TOKEN_ID, ACCESS_TOKEN, EXPIRES_IN, null);
        TokenExtension tokenExtension = new TokenExtension(BASE_URL, EXCHANGE_URL, null, true);

        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.storeMasterTokenData(exchangeTokenResponse,
                tokenExtension);
    }

    @Test(expectedExceptions = IllegalArgumentException.class,
            expectedExceptionsMessageRegExp = "Master token scope with the 'artifactory-cold' scope does not exist")
    public void testStoreMasterTokenDataWithNonColdScope() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        ExchangeTokenResponse exchangeTokenResponse =
                new ExchangeTokenResponse(TOKEN_ID, ACCESS_TOKEN, EXPIRES_IN, NON_COLD_SCOPE);
        TokenExtension tokenExtension = new TokenExtension(BASE_URL, EXCHANGE_URL, null, true);

        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.storeMasterTokenData(exchangeTokenResponse,
                tokenExtension);
    }

    @Test(expectedExceptions = IllegalArgumentException.class,
            expectedExceptionsMessageRegExp = "Can not parse the master token scope to get a cold namespace: " +
                    "namespace is empty")
    public void testStoreMasterTokenDataWithoutNamespace() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        ExchangeTokenResponse exchangeTokenResponse =
                new ExchangeTokenResponse(TOKEN_ID, ACCESS_TOKEN, EXPIRES_IN, SCOPE_WITHOUT_NAMESPACE);
        TokenExtension tokenExtension = new TokenExtension(BASE_URL, EXCHANGE_URL, null, true);

        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.storeMasterTokenData(exchangeTokenResponse,
                tokenExtension);
    }

    @Test(expectedExceptions = IllegalStateException.class,
            expectedExceptionsMessageRegExp = "Instance running in the Cold Storage mode")
    public void testStoreMasterTokenDataForNonWarmInstance() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(true);
        ExchangeTokenResponse exchangeTokenResponse =
                new ExchangeTokenResponse(TOKEN_ID, ACCESS_TOKEN, EXPIRES_IN, SCOPE_WITHOUT_NAMESPACE);
        TokenExtension tokenExtension = new TokenExtension(BASE_URL, EXCHANGE_URL, null, true);

        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.storeMasterTokenData(exchangeTokenResponse,
                tokenExtension);
    }

    @Test(expectedExceptions = IllegalStateException.class,
            expectedExceptionsMessageRegExp = "Instance running in the Cold Storage mode")
    public void testUpdateMasterTokenDataForNonWarmInstance() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(true);
        ExchangeTokenResponse exchangeTokenResponse =
                new ExchangeTokenResponse(TOKEN_ID, ACCESS_TOKEN, EXPIRES_IN, SCOPE_WITHOUT_NAMESPACE);
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.updateMasterTokenData(
                exchangeTokenResponse);
    }

    @Test
    public void testUpdateMasterTokenData() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        ExchangeTokenResponse exchangeTokenResponse =
                new ExchangeTokenResponse(TOKEN_ID, ACCESS_TOKEN, EXPIRES_IN, SCOPE_WITH_NAMESPACE);

        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.updateMasterTokenData(
                exchangeTokenResponse);

        ArgumentCaptor<String> tokenIdCapture = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<String> accessTokenCapture = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<String> tokenTypeCapture = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<Long> expireInCapture = ArgumentCaptor.forClass(Long.class);
        ArgumentCaptor<String> scopeCapture = ArgumentCaptor.forClass(String.class);

        verify(artifactRetentionAddon, times(1))
                .updateMasterToken(tokenIdCapture.capture(),
                        accessTokenCapture.capture(),
                        tokenTypeCapture.capture(),
                        expireInCapture.capture(),
                        scopeCapture.capture());
        assertEquals(TOKEN_ID, tokenIdCapture.getValue());
        assertEquals(ACCESS_TOKEN, accessTokenCapture.getValue());
        assertEquals("access_token", tokenTypeCapture.getValue());
        assertEquals(EXPIRES_IN, expireInCapture.getValue().longValue());
        assertEquals(SCOPE_WITH_NAMESPACE, scopeCapture.getValue());
    }

    @Test(expectedExceptions = IllegalStateException.class,
            expectedExceptionsMessageRegExp = "Instance running in the Cold Storage mode")
    public void testCreateMasterTokenExchangeRequestForNonWarmInstance() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(true);
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE
                .createMasterTokenExchangeRefreshRequest(SUBJECT);
    }

    @Test(expectedExceptions = IllegalStateException.class,
            expectedExceptionsMessageRegExp = "Can not create exchange request to refresh master token: namespace " +
                    "does not exist")
    public void testCreateMasterTokenExchangeRequestWhenNamespaceIsEmpty() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        when(artifactRetentionAddon.getJfrogColdStorageArchivingNamespace()).thenReturn("");
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE
                .createMasterTokenExchangeRefreshRequest(SUBJECT);
    }

    @Test
    public void testCreateMasterTokenExchangeRequestWhenNamespace() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        when(artifactRetentionAddon.getJfrogColdStorageArchivingNamespace()).thenReturn(NAMESPACE);
        ExchangeTokenRequest exchangeTokenRequest =
                ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE
                        .createMasterTokenExchangeRefreshRequest(
                                SUBJECT);
        assertEquals(SUBJECT, exchangeTokenRequest.getSubject());
        assertEquals(COLD_USE_CASE, exchangeTokenRequest.getTarget());
    }

    @Test(expectedExceptions = IllegalStateException.class,
            expectedExceptionsMessageRegExp = "Instance running in the Cold Storage mode")
    public void testCreateScopedTokenExchangeRequestForNonWarmInstance() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(true);
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.createScopedTokenExchangeRequest(SUBJECT);
    }

    @Test(expectedExceptions = IllegalStateException.class,
            expectedExceptionsMessageRegExp = "Can not create exchange request to refresh master token: namespace " +
                    "does not exist")
    public void testCreateScopedTokenExchangeRequestWhenNamespaceIsEmpty() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        when(artifactRetentionAddon.getJfrogColdStorageArchivingNamespace()).thenReturn("");
        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE
                .createMasterTokenExchangeRefreshRequest(SUBJECT);
    }

    @Test
    public void testCreateScopedTokenExchangeRequestWhenNamespace() {
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        when(artifactRetentionAddon.getJfrogColdStorageArchivingNamespace()).thenReturn(NAMESPACE);
        ExchangeTokenRequest exchangeTokenRequest =
                ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE.createScopedTokenExchangeRequest(
                        SUBJECT);
        assertEquals(SUBJECT, exchangeTokenRequest.getSubject());
        assertEquals(LIMITED_COLD_USE_CASE, exchangeTokenRequest.getTarget());
    }

    private String buildExtensionJson() {
        return JsonUtils.getInstance().writeValueAsString(
                new TokenExtension(BASE_URL, EXCHANGE_URL, PAIRING_URL, true));
    }

    private String buildOnlyRevocableExtensionJson() {
        return JsonUtils.getInstance().writeValueAsString(new TokenExtension(null, null, null, true));
    }
}