package org.artifactory.api.rest.common.model.continues.util;

import org.apache.commons.collections.CollectionUtils;
import org.artifactory.api.rest.common.model.continues.ContinueIndexPage;
import org.artifactory.api.rest.common.model.continues.ContinueResult;
import org.artifactory.api.rest.common.model.continues.FetchFunction;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.mockito.ArgumentMatchers;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
import org.testng.collections.Lists;

import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static java.util.List.of;
import static org.mockito.Mockito.when;

/**
 * @author Omri Ziv
 */
public class PagingUtilsTest {

    @Mock
    ArtifactoryHome artifactoryHome;
    @Mock
    ArtifactorySystemProperties artifactorySystemProperties;

    private AutoCloseable autoCloseable;

    @BeforeMethod
    public void setup() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        ArtifactoryHome.bind(artifactoryHome);
        when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        when(artifactorySystemProperties.getLongProperty(ArgumentMatchers.any(ConstantValues.class)))
                .thenAnswer(invocationOnMock -> {
                    ConstantValues constantValue = invocationOnMock.getArgument(0);
                    return Long.parseLong(constantValue.getDefValue().trim());
                });
    }

    @AfterMethod
    public void cleanup() throws Exception {
        ArtifactoryHome.unbind();
        autoCloseable.close();
    }

    private final List<FetchFunction<String>> functions =
            of(
                    new FetchFunction<>(this::firstFunc, 3L),
                    new FetchFunction<>(this::secondFunc, 3L),
                    new FetchFunction<>(this::thirdFunc, 0L),
                    new FetchFunction<>(this::forthFunc, 4L),
                    new FetchFunction<>(this::fifthFunc, 2L));

    private final List<FetchFunction<String>> functionsWithBlanks =
            of(
                    new FetchFunction<>(this::firstFuncWithBlanks, 3L),
                    new FetchFunction<>(this::secondFuncWithBlanks, 3L),
                    new FetchFunction<>(this::thirdFunc, 0L),
                    new FetchFunction<>(this::forthFuncWithBlanks, 4L),
                    new FetchFunction<>(this::fifthFuncWithBlanks, 2L));

    @Test
    public void testGetPagingFromMultipleFunctions() {
        ContinueResult<String> continueResult = org.artifactory.api.rest.common.model.continues.util.PagingUtils
                .getPagingFromMultipleFunctions(0, 2, functions);
        List<String> pageableResult = continueResult.getData();

        Assert.assertEquals(pageableResult.size(), 2);
        Assert.assertEquals(pageableResult, of("1", "2"));
    }

    @Test
    public void testGetPagingFromMultipleFunctionsSecondArray() {
        ContinueResult<String> continueResult = PagingUtils.getPagingFromMultipleFunctions(3, 3, functions);
        List<String> pageableResult = continueResult.getData();

        Assert.assertEquals(pageableResult.size(), 3);
        Assert.assertEquals(pageableResult, of("4", "5", "6"));
    }

    @Test
    public void testGetPagingFromMultipleFunctionsFirstAndSecondArray() {
        ContinueResult<String> continueResult = PagingUtils.getPagingFromMultipleFunctions(0, 5, functions);
        List<String> pageableResult = continueResult.getData();

        Assert.assertEquals(pageableResult.size(), 5);
        Assert.assertEquals(pageableResult, of("1", "2", "3", "4", "5"));
    }

    @Test
    public void testGetPagingFromMultipleFunctionsSecondAndThirdAndForthArray() {
        ContinueResult<String> continueResult = PagingUtils.getPagingFromMultipleFunctions(4, 4, functions);
        List<String> pageableResult = continueResult.getData();

        Assert.assertEquals(pageableResult.size(), 4);
        Assert.assertEquals(of("5", "6", "7", "8"), pageableResult);
    }

    @Test
    public void testGetPagingFromMultipleFunctionsOverflowGettingAll() {
        ContinueResult<String> continueResult = PagingUtils.getPagingFromMultipleFunctions(0, 100, functions);
        List<String> pageableResult = continueResult.getData();
        Assert.assertEquals(pageableResult.size(), 12);
        List<String> expected = IntStream.range(1, 13).mapToObj(Integer::toString).collect(Collectors.toList());
        Assert.assertEquals(pageableResult, expected);
    }

    @Test
    public void testGetZeroPagingFromMultipleFunctions() {
        ContinueResult<String> continueResult = PagingUtils.getPagingFromMultipleFunctions(0, 0, functions);
        List<String> pageableResult = continueResult.getData();

        Assert.assertEquals(pageableResult.size(), 0);
        Assert.assertEquals(pageableResult, Lists.newArrayList());
    }

    @Test
    public void testGetPagingFromMultipleFunctionsContinuePageObj() {
        ContinueIndexPage continueIndex = new ContinueIndexPage();
        continueIndex.setContinueIndex(0L);
        continueIndex.setLimit(2L);
        ContinueResult<String> continueResult = PagingUtils.getPagingFromMultipleFunctions(continueIndex, functions);
        List<String> pageableResult = continueResult.getData();
        Assert.assertEquals(pageableResult.size(), 2);
        Assert.assertEquals(pageableResult, of("1", "2"));
    }

    @Test
    public void testGetPagingFromMultipleFunctionsContinuePageObjDefalt() {
        ContinueIndexPage continueIndex = new ContinueIndexPage();
        continueIndex.setContinueIndex(0L);
        ContinueResult<String> continueResult = org.artifactory.api.rest.common.model.continues.util.PagingUtils
                .getPagingFromMultipleFunctions(continueIndex, functions);
        List<String> pageableResult = continueResult.getData();
        Assert.assertEquals(pageableResult.size(), 12);
        List<String> expected = IntStream.range(1, 13).mapToObj(Integer::toString).collect(Collectors.toList());
        Assert.assertEquals(pageableResult, expected);
    }

    @Test
    public void testGetPagingFromMultipleFunctionsArraysWithBlanks() {
        ContinueResult<String> continueResult = PagingUtils.getPagingFromMultipleFunctions(4, 4, functionsWithBlanks);
        List<String> pageableResult = continueResult.getData();
        Assert.assertEquals(pageableResult.size(), 4);
        Assert.assertEquals(pageableResult, of("5", "6", "8", "10"));
        Assert.assertEquals(continueResult.getContinueState(), "10");
    }

    @Test
    public void testGetPagingFromMultipleFunctionsArraysWithBlanksFromBeginning() {
        ContinueResult<String> continueResult = PagingUtils.getPagingFromMultipleFunctions(0, 5, functionsWithBlanks);
        List<String> pageableResult = continueResult.getData();

        Assert.assertEquals(pageableResult.size(), 5);
        Assert.assertEquals(pageableResult, of("1", "2", "5", "6", "8"));
        Assert.assertEquals(continueResult.getContinueState(), "8");
    }

    @Test
    public void testGetPagingContinuation() {
        ContinueResult<String> continueResult = PagingUtils.getPagingFromMultipleFunctions(0, 2, functionsWithBlanks);
        List<String> pageableResult = continueResult.getData();
        continueResult = PagingUtils
                .getPagingFromMultipleFunctions(getNextContinueState(continueResult), 3, functionsWithBlanks);
        pageableResult.addAll(continueResult.getData());
        continueResult = PagingUtils
                .getPagingFromMultipleFunctions(getNextContinueState(continueResult), 6, functionsWithBlanks);
        pageableResult.addAll(continueResult.getData());
        Assert.assertEquals(pageableResult.size(), 7);
        Assert.assertEquals(pageableResult, of("1", "2", "5", "6", "8", "10", "11"));
        Assert.assertNull(continueResult.getContinueState());
    }

    private int getNextContinueState(ContinueResult<String> continueResult) {

        return Integer.parseInt(continueResult.getContinueState());
    }


    private ContinueResult<String> firstFunc(int skip, int limit) {
        List<String> original = of("1", "2", "3");
        return getStringContinueResult(skip, limit, original);

    }

    private ContinueResult<String> secondFunc(int skip, int limit) {
        List<String> original = of("4", "5", "6");
        return getStringContinueResult(skip, limit, original);

    }

    private ContinueResult<String> thirdFunc(int skip, int limit) {
        List<String> original = Lists.newArrayList();
        return getStringContinueResult(skip, limit, original);
    }

    private ContinueResult<String> forthFunc(int skip, int limit) {
        List<String> original = of("7", "8", "9", "10");
        return getStringContinueResult(skip, limit, original);

    }

    private ContinueResult<String> fifthFunc(int skip, int limit) {
        List<String> original = of("11", "12");
        return getStringContinueResult(skip, limit, original);

    }

    private ContinueResult<String> firstFuncWithBlanks(int skip, int limit) {
        List<String> original = of("1", "2", "X");
        return getStringContinueResult(skip, limit, original);
    }

    private ContinueResult<String> secondFuncWithBlanks(int skip, int limit) {
        List<String> original = of("X", "5", "6");
        return getStringContinueResult(skip, limit, original);
    }

    private ContinueResult<String> forthFuncWithBlanks(int skip, int limit) {
        List<String> original = of("X", "8", "X", "10");
        return getStringContinueResult(skip, limit, original);
    }

    private ContinueResult<String> fifthFuncWithBlanks(int skip, int limit) {
        List<String> original = of("11", "X");
        return getStringContinueResult(skip, limit, original);
    }

    private ContinueResult<String> getStringContinueResult(int skip, int limit, List<String> original) {
        List<String> list = original.stream().skip(skip).filter(this::nonX).limit(limit).collect(Collectors.toList());
        int continueState = CollectionUtils.isNotEmpty(list) ? original.indexOf(list.get(list.size() - 1)) + 1 : 0;
        return new ContinueResult<>(continueState + "", list);
    }


    private boolean nonX(String item) {
        return !"X".equals(item);
    }

}
