package org.artifactory.storage.db.binstore.service;

import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.checksum.ChecksumType;
import org.artifactory.storage.BinaryInsertRetryException;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.binstore.dao.BinariesDao;
import org.artifactory.util.CollectionUtils;
import org.jetbrains.annotations.NotNull;
import org.jfrog.storage.binstore.BinariesInfo;
import org.jfrog.storage.binstore.common.BinaryElementImpl;
import org.jfrog.storage.binstore.db.entity.BinaryEntity;
import org.jfrog.storage.binstore.exceptions.BinaryStorageException;
import org.jfrog.storage.binstore.ifc.model.BinaryElement;
import org.jfrog.storage.binstore.service.BinariesDbService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import static org.artifactory.storage.db.binstore.util.BinaryServiceUtils.isDuplicatedEntryException;

/**
 * A business service to interact with the binaries table.
 *
 * @author Yossi Shaul
 */
@Service
public class BinariesDbServiceImpl implements BinariesDbService {
    private static final Logger log = LoggerFactory.getLogger(BinariesDbServiceImpl.class);

    private final BinariesDao binariesDao;

    public BinariesDbServiceImpl(@Autowired BinariesDao binariesDao) {
        this.binariesDao = binariesDao;
    }

    @Override
    public Optional<BinaryElement> findBinary(ChecksumType checksumType, String checksum) {
        try {
            BinaryEntity result = binariesDao.load(checksumType, checksum);
            return result != null ? Optional.of(toBinaryElement(result)) : Optional.empty();
        } catch (SQLException e) {
            throw new StorageException(
                    "Storage error loading checksum ( " + checksumType.name() + "): '" + checksum + "'", e);
        }
    }

    @Override
    public boolean exists(String sha1) {
        try {
            return binariesDao.exists(ChecksumType.sha1, sha1);
        } catch (SQLException sql) {
            log.debug("Failed existence check of path " + sha1, sql);
            throw new BinaryStorageException("Failed existence check of checksum " + sha1 + ": " + sql.getMessage(),
                    sql, HttpStatus.SC_INTERNAL_SERVER_ERROR);
        }
    }

    @Override
    @Nonnull
    public BinaryElement insertRecordInDb(String sha1, String sha2, String md5, long length) throws StorageException {
        validateBinaryInfo(sha1, sha2, md5, length);
        try {
            BinaryElement binary;
            Optional<BinaryElement> binaryOps = findBinary(ChecksumType.sha1, sha1);
            if (binaryOps.isEmpty()) {
                binary = new BinaryElementImpl(sha1, sha2, md5, length);
                createDataRecord(toBinaryEntity(binary));
            } else {
                binary = updateSha2IfMissing(sha2, binaryOps.get());
            }
            return binary;
        } catch (SQLException e) {
            throw new StorageException("Failed to insert new binary record: " + e.getMessage(), e);
        }
    }

    private BinaryElement updateSha2IfMissing(String sha2, BinaryElement binary) throws SQLException {
        // update a binary entity with sha2, this can happen if the binary was inserted before sha2 was enforced
        if (StringUtils.isBlank(binary.getSha2()) && StringUtils.isNotBlank(sha2)) {
            setSha2Checksum(binary.getSha1(), sha2);
            BinaryEntity binaryEntity = binariesDao.load(ChecksumType.sha1, binary.getSha1());
            if (binaryEntity == null) {
                throw new StorageException("Could not find just updated binary record: " + binary);
            }
            binary = toBinaryElement(binaryEntity);
        }
        return binary;
    }

    private void validateBinaryInfo(String sha1, String sha2, String md5, long length) {
        new BinaryEntityWithValidation(sha1, sha2, md5, length);
    }

    private void createDataRecord(BinaryEntity binaryEntity) throws SQLException {
        try {
            binariesDao.create(binaryEntity);
        } catch (SQLException e) {
            if (isDuplicatedEntryException(e)) {
                log.debug("Simultaneous insert of binary with sha1 {} detected, " +
                        "binary will be checked.", binaryEntity.getSha1(), e);
                throw new BinaryInsertRetryException(toBinaryElement(binaryEntity), e);
            } else {
                throw e;
            }
        }
    }

    @Override
    public long getStorageSize() {
        try {
            return binariesDao.getTotalSize();
        } catch (SQLException e) {
            throw new StorageException("Could not calculate total size due to " + e.getMessage(), e);
        }
    }

    @Override
    public BinariesInfo getCountAndTotalSize() {
        try {
            return binariesDao.getCountAndTotalSize();
        } catch (SQLException e) {
            throw new StorageException("Failed to retrieve binaries size", e);
        }
    }

    @Override
    public boolean setSha2Checksum(String sha1, String sha2) {
        try {
            return binariesDao.insertSha2(sha1, sha2);
        } catch (SQLException e) {
            throw new StorageException("Failed to set sha2 checksum " + sha2 + " for sha1 " + sha1);
        }
    }

    @Override
    public BinariesInfo getBinariesInfo() {
        try {
            return binariesDao.getCountAndTotalSize();
        } catch (SQLException e) {
            throw new StorageException("Could not calculate total size due to " + e.getMessage(), e);
        }
    }

    @Override
    public Collection<BinaryElement> findAllBinaries() {
        try {
            Collection<BinaryEntity> allBinaries = binariesDao.findAll();
            List<BinaryElement> result = Lists.newArrayListWithCapacity(allBinaries.size());
            result.addAll(allBinaries.stream().map(this::toBinaryElement).collect(Collectors.toList()));
            return result;
        } catch (SQLException e) {
            throw new StorageException("Could not retrieve all binary entries", e);
        }
    }

    @Override
    public int deleteEntry(String sha1ToDelete) {
        try {
            return binariesDao.deleteEntry(sha1ToDelete);
        } catch (SQLException e) {
            throw new StorageException("Could execute delete from binary store of " + sha1ToDelete);
        }
    }

    @NotNull
    @Override
    public Collection<BinaryElement> search(ChecksumType checksumType, Collection<String> validChecksums) {
        try {
            return binariesDao.search(checksumType, validChecksums)
                    .stream().map(this::toBinaryElement).collect(Collectors.toSet());
        } catch (SQLException e) {
            throw new StorageException("Could not search for checksums of type " + checksumType, e);
        }
    }

    @NotNull
    @Override
    public Set<BinaryElement> findBinaries(@Nullable Collection<String> checksums) {
        Set<BinaryElement> results = Sets.newHashSet();
        if (!CollectionUtils.isNullOrEmpty(checksums)) {
            for (ChecksumType checksumType : ChecksumType.BASE_CHECKSUM_TYPES) {
                Collection<String> checksumsOfType = getChecksumsofType(checksumType, checksums);
                results.addAll(search(checksumType, checksumsOfType));
            }
        }
        return results;
    }

    @Override
    public Collection<BinaryElement> findPotentialDeletion() {
        try {
            return binariesDao.findPotentialDeletion().stream().map(this::toBinaryElement).collect(Collectors.toList());
        } catch (SQLException e) {
            log.error("Could not find potential binaries to delete: " + e.getMessage());
            throw new StorageException("Could not find potential binaries to delete!", e);
        }
    }

    private Collection<String> getChecksumsofType(ChecksumType checksumType, Collection<String> checksums) {
        return checksums.stream().filter(checksumType::isValid).collect(Collectors.toSet());
    }

    private BinaryElement toBinaryElement(BinaryEntity bd) {
        return new BinaryElementImpl(bd.getSha1(), bd.getSha2(), bd.getMd5(), bd.getLength());
    }

    private BinaryEntity toBinaryEntity(BinaryElement be) {
        return new BinaryEntity(be.getSha1(), be.getSha2(), be.getMd5(), be.getLength());
    }

}
